from decimal import MAX_EMAX, MIN_EMIN, Decimal

from .types import FSharpRef

get_zero = Decimal(0)

get_one = Decimal(1)

get_minus_one = Decimal(-1)

get_max_value = MAX_EMAX

get_min_value = MIN_EMIN


def from_parts(low: int, mid: int, high: int, isNegative: bool, scale: int) -> Decimal:
    sign = -1 if isNegative else 1

    if low < 0:
        low = 0x100000000 + low

    if mid < 0:
        mid = 0xFFFFFFFF00000000 + mid + 1
    else:
        mid = mid << 32

    if high < 0:
        high = 0xFFFFFFFF0000000000000000 + high + 1
    else:
        high = high << 64

    value = Decimal((low + mid + high) * sign)
    if scale:
        dscale = Decimal(pow(10, scale))
        return value / dscale
    return value


def op_addition(x: Decimal, y: Decimal) -> Decimal:
    return x + y


def to_string(x: Decimal) -> str:
    return str(x)


def to_number(x: Decimal) -> float:
    return float(x)


def parse(string: str) -> Decimal:
    return Decimal(string)


def try_parse(string: str, def_value: FSharpRef[Decimal]) -> bool:
    try:
        def_value.contents = parse(string)
        return True
    except Exception:
        return False


def equals(a: Decimal, b: Decimal) -> bool:
    return a == b


__all__ = [
    "equals",
    "try_parse",
    "parse",
    "to_number",
    "to_string",
    "op_addition",
    "from_parts",
]



from __future__ import annotations
from typing import (Optional, Any)
from ...fable_library.option import (map, some)
from ...fs_hafas_python.context import Context
from ...fs_hafas_python.types_hafas_client import (Hint, Status)
from ...fs_hafas_python.types_raw_hafas_client import RawRem

default_hint : Hint = Hint("hint", None, None, "", None)

default_status : Status = Status("status", None, None, "", None)

def parse_hint(ctx: Context, h: RawRem) -> Optional[Any]:
    text : str
    match_value : Optional[str] = h.txt_n
    if match_value is None:
        text = ""
    
    else: 
        txt_n : str = match_value
        text = txt_n.strip()
    
    code : Optional[str] = h.code if (h.code != "") else None
    if h.type == "M":
        def mapping(s_1: str, ctx: Context=ctx, h: RawRem=h) -> str:
            return s_1.strip()
        
        return some(Status(default_status.type, code, map(mapping, h.txt_s), text, default_status.trip_id))
    
    elif "AI".find(h.type) >= 0:
        return some(Hint(default_hint.type, code, default_hint.summary, text, default_hint.trip_id))
    
    elif "DURNYQP".find(h.type) >= 0:
        return some(Status(default_status.type, code, default_status.summary, text, default_status.trip_id))
    
    else: 
        return None
    





from __future__ import annotations
from typing import (Any, TypeVar, Optional, Callable)
from .fsharp_collections import (ComparisonIdentity_Structural, HashIdentity_Structural)
from .system_text import (StringBuilder, StringBuilder__Append_Z721C83C5)
from .util import (equals, structural_hash, IComparer, IEqualityComparer, IDisposable, dispose, ignore)

_T = TypeVar("_T")

_A_ = TypeVar("_A_")

_B_ = TypeVar("_B_")

class ObjectExpr8(IEqualityComparer):
    def System_Collections_IEqualityComparer_Equals541DA560(self, x: Any=None, y: Any=None) -> bool:
        return equals(x, y)
    
    def System_Collections_IEqualityComparer_GetHashCode4E60E31B(self, x_1: Any=None) -> int:
        return structural_hash(x_1)
    

LanguagePrimitives_GenericEqualityComparer : IEqualityComparer = ObjectExpr8()

class ObjectExpr9(IEqualityComparer):
    def System_Collections_IEqualityComparer_Equals541DA560(self, x: Any=None, y: Any=None) -> bool:
        return equals(x, y)
    
    def System_Collections_IEqualityComparer_GetHashCode4E60E31B(self, x_1: Any=None) -> int:
        return structural_hash(x_1)
    

LanguagePrimitives_GenericEqualityERComparer : IEqualityComparer = ObjectExpr9()

def LanguagePrimitives_FastGenericComparer() -> IComparer[_T]:
    return ComparisonIdentity_Structural()


def LanguagePrimitives_FastGenericComparerFromTable() -> IComparer[_T]:
    return ComparisonIdentity_Structural()


def LanguagePrimitives_FastGenericEqualityComparer() -> IEqualityComparer[Any]:
    return HashIdentity_Structural()


def LanguagePrimitives_FastGenericEqualityComparerFromTable() -> IEqualityComparer[Any]:
    return HashIdentity_Structural()


def Operators_Failure(message: str) -> Exception:
    return Exception(message)


def Operators_FailurePattern(exn: Exception) -> Optional[str]:
    return str(exn)


def Operators_NullArg(x: str) -> _A_:
    raise Exception(x)


def Operators_Using(resource: IDisposable, action: Callable[[IDisposable], _A_]) -> _A_:
    try: 
        return action(resource)
    
    finally: 
        if equals(resource, None):
            pass
        
        else: 
            dispose(resource)
        
    


def Operators_Lock(_lockObj: _A_, action: Callable[[], _B_]) -> _B_:
    return action()


def ExtraTopLevelOperators_LazyPattern(input: Any) -> _A_:
    return input.Value


def PrintfModule_PrintFormatToStringBuilderThen(continuation: Callable[[], _A_], builder: StringBuilder, format: Any) -> _B_:
    def append(s: str, continuation: Callable[[], _A_]=continuation, builder: StringBuilder=builder, format: Any=format) -> _A_:
        ignore(StringBuilder__Append_Z721C83C5(builder, s))
        return continuation()
    
    return format.cont(append)


def PrintfModule_PrintFormatToStringBuilder(builder: StringBuilder, format: Any) -> _A_:
    def arrow_19(builder: StringBuilder=builder, format: Any=format) -> None:
        ignore()
    
    return PrintfModule_PrintFormatToStringBuilderThen(arrow_19, builder, format)





from __future__ import annotations
from typing import (Optional, List, Any, Tuple, Callable)
from ...fable_library.array import (try_find_index, choose, map, map_indexed, try_find, filter, exists as exists_1)
from ...fable_library.option import (to_array, default_arg, value as value_2, bind, map as map_1)
from ...fable_library.reflection import (TypeInfo, int32_type, bool_type, record_type)
from ...fable_library.seq import (exists, iterate)
from ...fable_library.types import Record
from ...fable_library.util import equals
from ...fs_hafas_python.context import (Context, Platform, CommonData, Profile__get_parseWhen, ParsedWhen, Profile__get_parsePlatform, Profile__get_parsePolyline, Profile__get_parseStopovers)
from ...fs_hafas_python.lib.transformations import (U2StationStop_FromSomeU3StationStopLocation, Default_Leg, Default_Location, Default_Alternative)
from ...fs_hafas_python.types_hafas_client import (StopOver, Leg, Line, FeatureCollection, Location, Cycle, Alternative)
from ...fs_hafas_python.types_raw_hafas_client import (RawMsg, RawSec, RawPltf, RawGis, RawJny, RawPoly, RawPolyG, RawCrd, RawFreq, RawStop)
from .common import (get_element_at_some, get_element_at, msg_lto_remarks)

def parse_platform(ctx: Context, platf_s: Optional[str]=None, platf_r: Optional[str]=None, cncl: Optional[bool]=None) -> Platform:
    planned : Optional[str] = platf_s
    prognosed : Optional[str] = platf_r
    def predicate(x: bool, ctx: Context=ctx, platf_s: Optional[str]=platf_s, platf_r: Optional[str]=platf_r, cncl: Optional[bool]=cncl) -> bool:
        return x
    
    if exists(predicate, to_array(cncl)):
        return Platform(None, planned, prognosed)
    
    else: 
        return Platform(default_arg(prognosed, planned), planned, None)
    


def expr_446() -> TypeInfo:
    return record_type("FsHafas.Parser.JourneyLeg.RemarkRange", [], RemarkRange, lambda: [("rem_x", int32_type), ("whole_leg", bool_type), ("from_index", int32_type), ("to_index", int32_type)])


class RemarkRange(Record):
    def __init__(self, rem_x: int, whole_leg: bool, from_index: int, to_index: int) -> None:
        super().__init__()
        self.rem_x = rem_x or 0
        self.whole_leg = whole_leg
        self.from_index = from_index or 0
        self.to_index = to_index or 0
    

RemarkRange_reflection = expr_446

def get_remark_range(msg: RawMsg, common: CommonData, stopovers: List[StopOver]) -> Optional[RemarkRange]:
    from_loc : Optional[Any] = U2StationStop_FromSomeU3StationStopLocation(get_element_at_some(msg.f_loc_x, common.locations))
    to_loc : Optional[Any] = U2StationStop_FromSomeU3StationStopLocation(get_element_at_some(msg.t_loc_x, common.locations))
    def predicate(s: StopOver, msg: RawMsg=msg, common: CommonData=common, stopovers: List[StopOver]=stopovers) -> bool:
        return equals(s.stop, from_loc)
    
    def predicate_1(s_1: StopOver, msg: RawMsg=msg, common: CommonData=common, stopovers: List[StopOver]=stopovers) -> bool:
        return equals(s_1.stop, to_loc)
    
    match_value : Tuple[Optional[int], Optional[int], Optional[int]] = (msg.rem_x, try_find_index(predicate, stopovers), try_find_index(predicate_1, stopovers))
    (pattern_matching_result, from_index_1, rem_x, to_index_1) = (None, None, None, None)
    if match_value[0] is not None:
        if match_value[1] is not None:
            if match_value[2] is not None:
                pattern_matching_result = 0
                from_index_1 = match_value[1]
                rem_x = match_value[0]
                to_index_1 = match_value[2]
            
            else: 
                pattern_matching_result = 1
            
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        return RemarkRange(rem_x, (to_index_1 == (len(stopovers) - 1)) if (from_index_1 == 0) else False, from_index_1, to_index_1)
    
    elif pattern_matching_result == 1:
        return None
    


def get_remark_ranges(msg_l: List[RawMsg], common_data: CommonData, stopovers: List[StopOver]) -> List[RemarkRange]:
    def chooser(x: Optional[RemarkRange]=None, msg_l: List[RawMsg]=msg_l, common_data: CommonData=common_data, stopovers: List[StopOver]=stopovers) -> Optional[RemarkRange]:
        return x
    
    def mapping(msg: RawMsg, msg_l: List[RawMsg]=msg_l, common_data: CommonData=common_data, stopovers: List[StopOver]=stopovers) -> Optional[RemarkRange]:
        return get_remark_range(msg, common_data, stopovers)
    
    return choose(chooser, map(mapping, msg_l, None), None)


def apply_remark_range(remark_range: RemarkRange, common_data: CommonData, stopover: StopOver) -> StopOver:
    def arrow_447(remark_range: RemarkRange=remark_range, common_data: CommonData=common_data, stopover: StopOver=stopover) -> List[Any]:
        match_value : Tuple[Optional[List[Any]], Optional[Optional[Any]]] = (stopover.remarks, get_element_at(remark_range.rem_x, common_data.hints))
        (pattern_matching_result, hint_2, remarks, hint_3) = (None, None, None, None)
        if match_value[0] is None:
            if match_value[1] is not None:
                if value_2(match_value[1]) is not None:
                    pattern_matching_result = 1
                    hint_3 = value_2(match_value[1])
                
                else: 
                    pattern_matching_result = 2
                
            
            else: 
                pattern_matching_result = 2
            
        
        elif match_value[1] is not None:
            pattern_matching_result = 0
            hint_2 = value_2(match_value[1])
            remarks = match_value[0]
        
        else: 
            pattern_matching_result = 2
        
        if pattern_matching_result == 0:
            return remarks
        
        elif pattern_matching_result == 1:
            return [value_2(hint_3)]
        
        elif pattern_matching_result == 2:
            return []
        
    
    return StopOver(stopover.stop, stopover.departure, stopover.departure_delay, stopover.prognosed_departure, stopover.planned_departure, stopover.departure_platform, stopover.prognosed_departure_platform, stopover.planned_departure_platform, stopover.arrival, stopover.arrival_delay, stopover.prognosed_arrival, stopover.planned_arrival, stopover.arrival_platform, stopover.prognosed_arrival_platform, stopover.planned_arrival_platform, arrow_447(), stopover.pass_by, stopover.cancelled)


def apply_remark_ranges(common_data: CommonData, stopovers: List[StopOver], remark_ranges: List[RemarkRange]) -> List[StopOver]:
    def mapping(i: int, s: StopOver, common_data: CommonData=common_data, stopovers: List[StopOver]=stopovers, remark_ranges: List[RemarkRange]=remark_ranges) -> StopOver:
        def predicate(r: RemarkRange, i: int=i, s: StopOver=s) -> bool:
            if (r.from_index <= i) if (not r.whole_leg) else False:
                return r.to_index >= i
            
            else: 
                return False
            
        
        match_value : Optional[RemarkRange] = try_find(predicate, remark_ranges)
        if match_value is None:
            return s
        
        else: 
            return apply_remark_range(match_value, common_data, s)
        
    
    return map_indexed(mapping, stopovers, None)


def parse_journey_leg(ctx: Context, pt: RawSec, date: str) -> Leg:
    leg : Leg = Default_Leg
    origin : Optional[Any] = get_element_at_some(pt.dep.loc_x, ctx.common.locations)
    destination : Optional[Any] = get_element_at_some(pt.arr.loc_x, ctx.common.locations)
    dep : ParsedWhen = Profile__get_parseWhen(ctx.profile)(ctx)(date)(pt.dep.d_time_s)(pt.dep.d_time_r)(pt.dep.d_tzoffset)(pt.dep.d_cncl)
    arr : ParsedWhen = Profile__get_parseWhen(ctx.profile)(ctx)(date)(pt.arr.a_time_s)(pt.arr.a_time_r)(pt.arr.a_tzoffset)(pt.arr.a_cncl)
    def match_platf_s(a_platf_s: Optional[str]=None, a_pltf_s: Optional[RawPltf]=None, ctx: Context=ctx, pt: RawSec=pt, date: str=date) -> Optional[str]:
        if a_platf_s is None:
            if a_pltf_s is not None:
                return a_pltf_s.txt
            
            else: 
                return None
            
        
        else: 
            return a_platf_s
        
    
    d_platf_s : Optional[str] = match_platf_s(pt.dep.d_platf_s, pt.dep.d_pltf_s)
    d_platf_r : Optional[str] = match_platf_s(pt.dep.d_platf_r, pt.dep.d_pltf_r)
    dep_pl : Platform = Profile__get_parsePlatform(ctx.profile)(ctx)(d_platf_s)(d_platf_r)(pt.dep.d_cncl)
    a_platf_s_1 : Optional[str] = match_platf_s(pt.arr.a_platf_s, pt.arr.a_pltf_s)
    a_platf_r : Optional[str] = match_platf_s(pt.arr.a_platf_r, pt.arr.a_pltf_r)
    arr_pl : Platform = Profile__get_parsePlatform(ctx.profile)(ctx)(a_platf_s_1)(a_platf_r)(pt.arr.a_cncl)
    if True if (True if (pt.type == "WALK") else (pt.type == "TRSF")) else (pt.type == "DEVI"):
        def arrow_448(ctx: Context=ctx, pt: RawSec=pt, date: str=date) -> Leg:
            distance : Optional[int]
            match_value : Optional[RawGis] = pt.gis
            distance = None if (match_value is None) else match_value.dist
            transfer : Optional[bool] = True if (pt.type == "TRSF") else (pt.type == "DEVI")
            return Leg(leg.trip_id, leg.origin, leg.destination, leg.departure, leg.planned_departure, leg.prognosed_arrival, leg.departure_delay, leg.departure_platform, leg.prognosed_departure_platform, leg.planned_departure_platform, leg.arrival, leg.planned_arrival, leg.prognosed_departure, leg.arrival_delay, leg.arrival_platform, leg.prognosed_arrival_platform, leg.planned_arrival_platform, leg.stopovers, leg.schedule, leg.price, leg.operator, leg.direction, leg.line, leg.reachable, leg.cancelled, True, leg.load_factor, distance, True, transfer, leg.cycle, leg.alternatives, leg.polyline, leg.remarks, leg.current_location)
        
        leg = arrow_448()
    
    if pt.type == "JNY":
        def action(jny: RawJny, ctx: Context=ctx, pt: RawSec=pt, date: str=date) -> None:
            nonlocal leg
            line : Optional[Line] = get_element_at(jny.prod_x, ctx.common.lines)
            polyline : Optional[FeatureCollection]
            match_value_1 : Optional[RawPoly] = jny.poly
            if match_value_1 is None:
                match_value_2 : Optional[RawPolyG] = jny.poly_g
                if match_value_2 is None:
                    polyline = None
                
                else: 
                    poly_g : RawPolyG = match_value_2
                    idx : int = poly_g.poly_xl[0] or 0
                    polyline = ctx.common.polylines[idx] if (idx < len(ctx.common.polylines)) else None
                
            
            else: 
                value : RawPoly = match_value_1
                polyline = Profile__get_parsePolyline(ctx.profile)(ctx)(value)
            
            stopovers : Optional[List[StopOver]] = Profile__get_parseStopovers(ctx.profile)(ctx)(jny.stop_l)(date) if ctx.opt.stopovers else None
            remark_ranges : List[RemarkRange]
            match_value_4 : Tuple[Optional[List[RawMsg]], Optional[List[StopOver]]] = (jny.msg_l, stopovers)
            (pattern_matching_result, msg_l, stopovers_1) = (None, None, None)
            if match_value_4[0] is not None:
                if match_value_4[1] is not None:
                    pattern_matching_result = 0
                    msg_l = match_value_4[0]
                    stopovers_1 = match_value_4[1]
                
                else: 
                    pattern_matching_result = 1
                
            
            else: 
                pattern_matching_result = 1
            
            if pattern_matching_result == 0:
                remark_ranges = get_remark_ranges(msg_l, ctx.common, stopovers_1)
            
            elif pattern_matching_result == 1:
                remark_ranges = []
            
            stopovers_with_remarks : Optional[List[StopOver]] = apply_remark_ranges(ctx.common, stopovers, remark_ranges) if (stopovers is not None) else None
            msg_l_2 : Optional[List[RawMsg]]
            match_value_5 : Optional[List[RawMsg]] = jny.msg_l
            def predicate_1(msg: RawMsg, jny: RawJny=jny) -> bool:
                def predicate(r: RemarkRange, msg: RawMsg=msg) -> bool:
                    if equals(r.rem_x, msg.rem_x):
                        return not r.whole_leg
                    
                    else: 
                        return False
                    
                
                return not exists_1(predicate, remark_ranges)
            
            msg_l_2 = None if (match_value_5 is None) else filter(predicate_1, match_value_5)
            current_location : Optional[Location]
            match_value_6 : Optional[RawCrd] = jny.pos
            if match_value_6 is None:
                current_location = None
            
            else: 
                pos : RawCrd = match_value_6
                current_location = Location(Default_Location.type, Default_Location.id, Default_Location.name, Default_Location.poi, Default_Location.address, pos.x / 1000000, pos.y / 1000000, Default_Location.altitude, Default_Location.distance)
            
            remarks : Optional[List[Any]] = msg_lto_remarks(ctx, msg_l_2) if ctx.opt.remarks else None
            cycle : Optional[Cycle]
            match_value_7 : Optional[RawFreq] = jny.freq
            if match_value_7 is None:
                cycle = None
            
            else: 
                freq : RawFreq = match_value_7
                match_value_8 : Tuple[Optional[int], Optional[int]] = (freq.min_c, freq.max_c)
                def arrow_449(jny: RawJny=jny) -> Optional[Cycle]:
                    max_c : int = match_value_8[1] or 0
                    min_c : int = match_value_8[0] or 0
                    return Cycle(min_c * 60, max_c * 60, freq.num_c)
                
                def arrow_450(jny: RawJny=jny) -> Optional[Cycle]:
                    min_c_1 : int = match_value_8[0] or 0
                    return Cycle(min_c_1 * 60, None, None)
                
                cycle = (arrow_449() if (match_value_8[1] is not None) else arrow_450()) if (match_value_8[0] is not None) else None
            
            def binder_4(freq_1: RawFreq, jny: RawJny=jny) -> Optional[List[Alternative]]:
                def mapping(array_2: List[RawJny], freq_1: RawFreq=freq_1) -> List[Alternative]:
                    def parse_alternative(a: RawJny, array_2: List[RawJny]=array_2) -> Alternative:
                        line_1 : Optional[Line] = get_element_at(a.prod_x, ctx.common.lines)
                        parsed_when : Optional[ParsedWhen]
                        match_value_9 : Optional[List[RawStop]] = a.stop_l
                        (pattern_matching_result_1, stop_l_1) = (None, None)
                        if match_value_9 is not None:
                            if len(match_value_9) > 0:
                                pattern_matching_result_1 = 0
                                stop_l_1 = match_value_9
                            
                            else: 
                                pattern_matching_result_1 = 1
                            
                        
                        else: 
                            pattern_matching_result_1 = 1
                        
                        if pattern_matching_result_1 == 0:
                            st0 : RawStop = stop_l_1[0]
                            parsed_when = Profile__get_parseWhen(ctx.profile)(ctx)(date)(st0.d_time_s)(st0.d_time_r)(st0.d_tzoffset)(st0.d_cncl)
                        
                        elif pattern_matching_result_1 == 1:
                            parsed_when = None
                        
                        def binder(v: ParsedWhen, a: RawJny=a) -> Optional[str]:
                            return v.when
                        
                        def binder_1(v_1: ParsedWhen, a: RawJny=a) -> Optional[str]:
                            return v_1.planned_when
                        
                        def binder_2(v_2: ParsedWhen, a: RawJny=a) -> Optional[str]:
                            return v_2.prognosed_when
                        
                        def binder_3(v_3: ParsedWhen, a: RawJny=a) -> Optional[int]:
                            return v_3.delay
                        
                        return Alternative(a.jid, a.dir_txt, Default_Alternative.location, line_1, Default_Alternative.stop, bind(binder, parsed_when), bind(binder_1, parsed_when), bind(binder_2, parsed_when), bind(binder_3, parsed_when), Default_Alternative.platform, Default_Alternative.planned_platform, Default_Alternative.prognosed_platform, Default_Alternative.remarks, Default_Alternative.cancelled, Default_Alternative.load_factor, Default_Alternative.provenance, Default_Alternative.previous_stopovers, Default_Alternative.next_stopovers, Default_Alternative.frames, Default_Alternative.polyline, Default_Alternative.current_trip_position, Default_Alternative.origin, Default_Alternative.destination)
                    
                    return map(parse_alternative, array_2, None)
                
                return map_1(mapping, freq_1.jny_l)
            
            alternatives : Optional[List[Alternative]] = bind(binder_4, jny.freq)
            leg = Leg(jny.jid, leg.origin, leg.destination, leg.departure, leg.planned_departure, leg.prognosed_arrival, leg.departure_delay, leg.departure_platform, leg.prognosed_departure_platform, leg.planned_departure_platform, leg.arrival, leg.planned_arrival, leg.prognosed_departure, leg.arrival_delay, leg.arrival_platform, leg.prognosed_arrival_platform, leg.planned_arrival_platform, stopovers_with_remarks, leg.schedule, leg.price, leg.operator, jny.dir_txt, line, jny.is_rchbl, leg.cancelled, leg.walking, leg.load_factor, leg.distance, leg.public, leg.transfer, cycle, alternatives, polyline, remarks, current_location)
        
        iterate(action, to_array(pt.jny))
    
    cancelled : Optional[bool] = default_arg(pt.dep.d_cncl, pt.arr.a_cncl)
    return Leg(leg.trip_id, origin, destination, dep.when, dep.planned_when, arr.prognosed_when, dep.delay, dep_pl.platform, dep_pl.prognosed_platform, dep_pl.planned_platform, arr.when, arr.planned_when, dep.prognosed_when, arr.delay, arr_pl.platform, arr_pl.prognosed_platform, arr_pl.planned_platform, leg.stopovers, leg.schedule, leg.price, leg.operator, leg.direction, leg.line, leg.reachable, cancelled, leg.walking, leg.load_factor, leg.distance, leg.public, leg.transfer, leg.cycle, leg.alternatives, leg.polyline, leg.remarks, leg.current_location)





from abc import abstractmethod
from typing import (Any, Protocol)

class IExports(Protocol):
    @abstractmethod
    def dumps(self, obj: Any) -> str:
        ...
    
    @abstractmethod
    def loads(self, __arg0: str) -> Any:
        ...
    




from typing import Any

from .types import FSharpRef


def try_parse(string: str, defValue: FSharpRef[bool]) -> bool:
    string = string.strip().capitalize()
    if string == "True":
        defValue.contents = True
        return True
    elif string == "False":
        defValue.contents = False
        return True

    return False


def parse(value: Any) -> bool:
    def_value: FSharpRef[bool] = FSharpRef(False)
    if try_parse(value, def_value):
        return def_value.contents

    raise ValueError(f"String '{value}' was not recognized as a valid Boolean.")



from __future__ import annotations
from abc import abstractmethod
from typing import (Callable, Protocol)

class ITimer(Protocol):
    @abstractmethod
    def Create(self, __arg0: float, __arg1: Callable[[], None]) -> ITimer:
        ...
    
    @abstractmethod
    def cancel(self) -> None:
        ...
    
    @property
    @abstractmethod
    def daemon(self) -> bool:
        ...
    
    @daemon.setter
    @abstractmethod
    def daemon(self, __arg0: bool) -> None:
        ...
    
    @abstractmethod
    def start(self) -> None:
        ...
    




from typing import Any, Optional, Tuple

from .types import FSharpRef


def from_bits(lowBits: int, highBits: int, unsigned: bool) -> int:
    ret = lowBits + (highBits << 32)
    if not unsigned and ret > 0x7FFFFFFFFFFFFFFF:
        return ret - 0x10000000000000000

    return ret


def from_int(value: int, unsigned: bool = False) -> int:
    if unsigned and value < 0:
        return value + 0x10000000000000000
    return value


def from_value(value: Any, unsigned: bool = False) -> int:
    value = int(value)
    if unsigned and value < 0:
        return value + 0x10000000000000000
    elif not unsigned and value > 9223372036854775807:
        return value - 0x10000000000000000

    return value


def from_number(value: int, unsigned: bool) -> int:
    return int(value)


def to_number(value: int) -> float:
    return float(value)


def from_integer(value: int, unsigned: Optional[bool] = None, kind: Optional[int] = None) -> int:
    if unsigned and value < 0:
        return value + 0x10000000000000000
    elif not unsigned and value > 9223372036854775807:
        return value - 0x10000000000000000
    return value


def op_left_shift(self: int, numBits: int):
    return self << numBits


def op_addition(a: int, b: int) -> int:
    return a + b


def op_multiply(a: int, b: int) -> int:
    return a * b


def op_unary_negation(value: int) -> int:
    return -value


def get_range(unsigned: bool) -> Tuple[int, int]:
    return (0, 18446744073709551615) if unsigned else (-9223372036854775808, 9223372036854775807)


AllowHexSpecifier = 0x00000200


def parse(string: str, style: int, unsigned: bool, bitsize: int, radix: int = 10) -> int:
    # const res = isValid(str, style, radix);
    if style & AllowHexSpecifier or string.startswith("0x"):
        radix = 16
    elif string.startswith("0b"):
        radix = 2
    elif string.startswith("0o"):
        radix = 8

    try:
        v = int(string, base=radix)
    except Exception:
        raise ValueError("Input string was not in a correct format.")

    (umin, umax) = get_range(True)
    if not unsigned and radix != 10 and v >= umin and v <= umax:
        mask = 1 << (bitsize - 1)
        if v & mask:  # Test if negative
            v = v - (mask << 1)

    (min, max) = get_range(unsigned)
    if v >= min and v <= max:
        return v

    raise ValueError("Input string was not in a correct format.")


def try_parse(string: str, style: int, unsigned: bool, bitsize: int, defValue: FSharpRef[int]) -> bool:
    try:
        defValue.contents = parse(string, style, unsigned, bitsize)
        return True
    except Exception:
        return False


def to_string(x: int) -> str:
    return str(x)


def to_int(value: int) -> int:
    if value > 9223372036854775807:
        return value - 0x10000000000000000
    return value


def compare(a: int, b: int) -> int:
    return 0 if a == b else -1 if a < b else 1


def equals(a: int, b: int) -> bool:
    return a == b


long = int
